/*
* Copyright (c) 2008-2020, NVIDIA CORPORATION. All rights reserved.
*
* NVIDIA CORPORATION and its licensors retain all intellectual property
* and proprietary rights in and to this software, related documentation
* and any modifications thereto. Any use, reproduction, disclosure or
* distribution of this software and related documentation without an express
* license agreement from NVIDIA CORPORATION is strictly prohibited. 
*/

#include <waveworks/GFSDK_WaveWorks.h>
#include <applib/DeviceManager.h>

#include "./main_state.h"
#include "./ui_renderer.h"
#include "./sky_renderer.h"
#include "./ocean_renderer.h"

#ifndef SAFE_DELETE
#define SAFE_DELETE(p)       { if (p) { delete (p);     (p) = nullptr; } }
#endif
#ifndef SAFE_DELETE_ARRAY
#define SAFE_DELETE_ARRAY(p) { if (p) { delete[] (p);   (p) = nullptr; } }
#endif
#ifndef SAFE_RELEASE
#define SAFE_RELEASE(p)      { if (p) { (p)->Release(); (p) = nullptr; } }
#endif

// Global variables
MainState									g_mainState;
DeviceManager*								g_pDeviceManager = nullptr;
UIRenderer*									g_pUIRenderer = nullptr;
SkyRenderer*						        g_pSkyRenderer = nullptr;
OceanRenderer*						        g_pOceanRenderer = nullptr;
//InteractiveSimulation*						g_pInteractiveSimulation = nullptr;

#if USE_DX11
    #define API_STRING "DX11"
#elif USE_DX12
    #define API_STRING "DX12"
#elif USE_VK
    #define API_STRING "Vulkan"
#endif

static const char *windowTitle = "NVIDIA GameWorks Demo (" API_STRING ")";

int main(int argc, char **argv)
{
	_CrtSetDbgFlag ( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );  
	
	// Initializing device manager
	g_pDeviceManager = DeviceManager::Create();

	// Initializing the WaveWorks library
	GFSDK_WaveWorks_Init(nullptr, GFSDK_WAVEWORKS_API_GUID);

	// Initializing the wind simulation parameters & settings
	g_mainState.OceanWindSimulationParameters.time_scale = 1.0f;
	g_mainState.OceanWindSimulationParameters.lateral_multiplier = 1.f;
	g_mainState.OceanWindSimulationParameters.uv_warping_amplitude = 0.03f;
	g_mainState.OceanWindSimulationParameters.uv_warping_frequency = 2.f;

	g_mainState.OceanWindSimulationParameters.base_wind_direction = { 1.0f, 0.0f };
	g_mainState.OceanWindSimulationParameters.base_wind_speed = 4.7f;
	g_mainState.OceanWindSimulationParameters.base_wind_distance = 0.1f;
	g_mainState.OceanWindSimulationParameters.base_wind_dependency = 1.0f;
	g_mainState.OceanWindSimulationParameters.base_spectrum_peaking = 3.3f;
	g_mainState.OceanWindSimulationParameters.base_small_waves_cutoff_length = 0.0f;
	g_mainState.OceanWindSimulationParameters.base_small_waves_cutoff_power = 0.0f;
	g_mainState.OceanWindSimulationParameters.base_amplitude_multiplier = 1.0f;

	g_mainState.OceanWindSimulationParameters.swell_wind_direction = { 0.0f, 1.0f };
	g_mainState.OceanWindSimulationParameters.swell_wind_speed = 1.5f;
	g_mainState.OceanWindSimulationParameters.swell_wind_distance = 520.0f;
	g_mainState.OceanWindSimulationParameters.swell_wind_dependency = 1.0f;
	g_mainState.OceanWindSimulationParameters.swell_spectrum_peaking = 10.0f;
	g_mainState.OceanWindSimulationParameters.swell_small_waves_cutoff_length = 60.0f;
	g_mainState.OceanWindSimulationParameters.swell_small_waves_cutoff_power = 1.0f;
	g_mainState.OceanWindSimulationParameters.swell_amplitude_multiplier = 1.0f;

	g_mainState.OceanWindSimulationParameters.foam_whitecaps_threshold = 0.5f;
	g_mainState.OceanWindSimulationParameters.foam_dissipation_speed = 0.6f;
	g_mainState.OceanWindSimulationParameters.foam_falloff_speed = 0.985f;
	g_mainState.OceanWindSimulationParameters.foam_generation_amount = 0.12f;
	g_mainState.OceanWindSimulationParameters.foam_generation_threshold = 0.37f;

	g_mainState.OceanWindSimulationSettings.simulation_period = 1000.0f;
	g_mainState.OceanWindSimulationSettings.simulation_api = GFSDK_WaveWorks_Simulation_API_Compute;
	g_mainState.OceanWindSimulationSettings.detail_level = GFSDK_WaveWorks_Simulation_DetailLevel_Extreme;
	g_mainState.OceanWindSimulationSettings.enable_CPU_driven_displacement_calculation = true;
	g_mainState.OceanWindSimulationSettings.enable_GPU_driven_displacement_calculation = true;
	g_mainState.OceanWindSimulationSettings.num_readback_FIFO_entries = g_mainState.ReadbackArchiveSize;
	g_mainState.OceanWindSimulationSettings.CPU_simulation_threading_model = GFSDK_WaveWorks_Simulation_CPU_Threading_Model_Automatic;
	g_mainState.OceanWindSimulationSettings.use_Beaufort_scale = true;
	g_mainState.OceanWindSimulationSettings.num_GPUs = 1;
	g_mainState.OceanWindSimulationSettings.enable_GPU_timers = true;
	g_mainState.OceanWindSimulationSettings.enable_CPU_timers = true;

	// Initializing the local simulation parameters & settings
	g_mainState.OceanLocalSimulationParameters.amplitude_multiplier = 1.0f;
	g_mainState.OceanLocalSimulationParameters.lateral_multiplier = 1.0f;

	g_mainState.OceanLocalSimulationParameters.foam_whitecaps_threshold = 0.3f;
	g_mainState.OceanLocalSimulationParameters.foam_dissipation_speed = 0.3f;
	g_mainState.OceanLocalSimulationParameters.foam_falloff_speed = 0.6f;
	g_mainState.OceanLocalSimulationParameters.foam_generation_amount = 0.2f;
	g_mainState.OceanLocalSimulationParameters.foam_generation_threshold = 0.6f;

	g_mainState.OceanLocalSimulationSettings.simulation_domain_center.x = 0;
	g_mainState.OceanLocalSimulationSettings.simulation_domain_center.y = 0;
	g_mainState.OceanLocalSimulationSettings.simulation_api = GFSDK_WaveWorks_Simulation_API_Compute;
	g_mainState.OceanLocalSimulationSettings.simulation_domain_worldspace_size = 200.0f;
	g_mainState.OceanLocalSimulationSettings.simulation_domain_grid_size = 512;
	g_mainState.OceanLocalSimulationSettings.enable_CPU_driven_displacement_calculation = true;
	g_mainState.OceanLocalSimulationSettings.enable_GPU_driven_displacement_calculation = true;
	g_mainState.OceanLocalSimulationSettings.CPU_simulation_threading_model = GFSDK_WaveWorks_Simulation_CPU_Threading_Model_Automatic;
	g_mainState.OceanLocalSimulationSettings.num_GPUs = 1;
	g_mainState.OceanLocalSimulationSettings.enable_GPU_timers = true;
	g_mainState.OceanLocalSimulationSettings.enable_CPU_timers = true;

	// Resetting the stats
	memset(&g_mainState.OceanWindSimulationStats, 0, sizeof(g_mainState.OceanWindSimulationStats));
	memset(&g_mainState.OceanWindSimulationStatsFiltered, 0, sizeof(g_mainState.OceanWindSimulationStatsFiltered));

	memset(&g_mainState.OceanLocalSimulationStats, 0, sizeof(g_mainState.OceanLocalSimulationStats));
	memset(&g_mainState.OceanLocalSimulationStatsFiltered, 0, sizeof(g_mainState.OceanLocalSimulationStatsFiltered));

	g_mainState.bForceKick = true;

	// Initializing quadtree parameters
	g_mainState.OceanQuadtreeParameters.min_patch_length = 5.0f;
	g_mainState.OceanQuadtreeParameters.max_edge_length = 10.0f;
	g_mainState.OceanQuadtreeParameters.cell_count = 64;
	g_mainState.OceanQuadtreeParameters.mean_sea_level = 0.0f;
	g_mainState.OceanQuadtreeParameters.max_LOD_number = 15;
	g_mainState.OceanQuadtreeParameters.geomorphing_degree = 1.0f;
	g_mainState.OceanQuadtreeParameters.generate_diamond_pattern = false;

	// Creating device and swap chain
	DeviceCreationParameters deviceParams;
	deviceParams.backBufferWidth = 1280;
	deviceParams.backBufferHeight = 800;
#if USE_VK
	deviceParams.swapChainFormat = nvrhi::Format::BGRA8_UNORM;
#else
	deviceParams.swapChainFormat = nvrhi::Format::RGBA8_UNORM;
#endif
	deviceParams.swapChainSampleCount = 1;
	deviceParams.swapChainSampleQuality = 0;
	deviceParams.swapChainBufferCount = 2;
	deviceParams.startFullscreen = false;
	deviceParams.enableDebugRuntime = false;
	deviceParams.enableNvrhiValidationLayer = false;

	

	if (!g_pDeviceManager->CreateWindowDeviceAndSwapChain(deviceParams, windowTitle))
	{
#if WIN32
		MessageBoxA(NULL, (LPCSTR)L"Cannot initialize the " API_STRING " device with the requested parameters", (LPCSTR)L"Error", MB_OK | MB_ICONERROR);
#else
		fprintf(stderr, "Cannot initialize the " API_STRING "GPU device with the requested parameters");
#endif
		return 1;
	}
	
	// Creating wind waves simulation
	GFSDK_WaveWorks_Wind_Waves_Simulation_CreateDirectX12(
		g_pDeviceManager->GetDevice()->getNativeObject(nvrhi::ObjectTypes::D3D12_Device), 
		g_pDeviceManager->GetDevice()->getNativeObject(nvrhi::ObjectTypes::D3D12_GraphicsCommandQueue), 
		true, 
		g_mainState.OceanWindSimulationSettings, 
		g_mainState.OceanWindSimulationParameters, 
		&(g_mainState.hOceanWindSimulation));

	// Creating local waves simulation
	GFSDK_WaveWorks_Local_Waves_Simulation_CreateDirectX12(
		g_pDeviceManager->GetDevice()->getNativeObject(nvrhi::ObjectTypes::D3D12_Device), 
		g_pDeviceManager->GetDevice()->getNativeObject(nvrhi::ObjectTypes::D3D12_GraphicsCommandQueue),
		true, g_mainState.OceanLocalSimulationSettings, 
		g_mainState.OceanLocalSimulationParameters, 
		&(g_mainState.hOceanLocalSimulation));

	// Creating quadtree
	GFSDK_WaveWorks_Quadtree_Create(g_mainState.OceanQuadtreeParameters, &g_mainState.hOceanQuadtree);

	// Initializing render passes
	g_pUIRenderer = new UIRenderer(g_pDeviceManager, &g_mainState);
	g_pSkyRenderer = new SkyRenderer(g_pDeviceManager, &g_mainState);
	g_pOceanRenderer = new OceanRenderer(g_pDeviceManager, &g_mainState);

	g_pDeviceManager->AddRenderPassToFront(g_pOceanRenderer);
	g_pDeviceManager->AddRenderPassToBack(g_pSkyRenderer);
	g_pDeviceManager->AddRenderPassToBack(g_pUIRenderer);
	g_pOceanRenderer->ResetCameraView();

	g_pUIRenderer->Init();
	g_pSkyRenderer->Init();
	g_pOceanRenderer->Init();


	// Main loop
	g_pDeviceManager->RunMessageLoop();

	// Destroying renderers
	SAFE_DELETE(g_pUIRenderer);

	g_pSkyRenderer->Shutdown();
	SAFE_DELETE(g_pSkyRenderer);

	g_pOceanRenderer->Shutdown();
	SAFE_DELETE(g_pOceanRenderer);

	// Destroying WaveWorks
	GFSDK_WaveWorks_Wind_Waves_Simulation_Destroy(g_mainState.hOceanWindSimulation);
	GFSDK_WaveWorks_Local_Waves_Simulation_Destroy(g_mainState.hOceanLocalSimulation);
	GFSDK_WaveWorks_Quadtree_Destroy(g_mainState.hOceanQuadtree);

	GFSDK_WaveWorks_Release();

	// Destroying device manager
	g_pDeviceManager->Shutdown();
	SAFE_DELETE(g_pDeviceManager);

	return 0;
}
